#include <queue>
#include <stack>
#include <fstream>
#include <iostream>

#include "BvhLoader.h"
#include "gzstream.h"


namespace CHCDemoEngine
{

using namespace std;


#define TYPE_INTERIOR -2
#define TYPE_LEAF -3



BvhNode *BvhLoader::LoadNextNode(igzstream &stream, BvhInterior *parent)
{
	int nodeType;
	stream.read(reinterpret_cast<char *>(&nodeType), sizeof(int));

	BvhNode *node;

	if (nodeType == TYPE_LEAF)
		node = new BvhLeaf(parent);
	else if (nodeType == TYPE_INTERIOR)
		node = new BvhInterior(parent);
	else
		cerr << "error: wrong node type: " << nodeType << endl;
		

	Vector3 bMin, bMax;

	stream.read(reinterpret_cast<char *>(&(node->mFirst)), sizeof(int));
	stream.read(reinterpret_cast<char *>(&(node->mLast)), sizeof(int));

	stream.read(reinterpret_cast<char *>(&bMin), sizeof(Vector3));
	stream.read(reinterpret_cast<char *>(&bMax), sizeof(Vector3));

	node->mBox = AxisAlignedBox3(bMin, bMax);
	node->mArea = node->mBox.SurfaceArea();

	//cout << "box: " << node->mBox << " area: " << node->mArea << endl;
	
	return node;
}


Bvh *BvhLoader::Load(const string &filename,  
					 const SceneEntityContainer &entities)
{
	queue<BvhNode *> tQueue;
	igzstream stream(filename.c_str());

	if (!stream.is_open()) return NULL;

	cout << "loading bvh" << endl;

	Bvh *bvh = new Bvh(entities);

	BvhNode *root = LoadNextNode(stream, NULL);

#if 0

	bvh->mRoot = root;

#else
	// copy root and set new one for dynamic objects
	BvhInterior *newRoot = new BvhInterior(NULL);
	
	newRoot->mBox = root->mBox;
	newRoot->mFirst = root->mFirst;
	newRoot->mLast = root->mLast;

	// create 'dynamic' leaf which basically is a container
	// for all dynamic objects
	BvhLeaf *dynamicLeaf = new BvhLeaf(newRoot);
	dynamicLeaf->mBox = root->mBox;

	newRoot->mFront = root;
	root->mParent = newRoot;

	newRoot->mBack = dynamicLeaf;

	bvh->mRoot = newRoot;
#endif

	tQueue.push(bvh->mRoot);
	bvh->mNumNodes = 1;

	while(!tQueue.empty())
	{
		BvhNode *node = tQueue.front();
		tQueue.pop();

		if (!node->IsLeaf())
		{
			bvh->mNumNodes += 2;

			BvhInterior *interior = static_cast<BvhInterior *>(node);

			BvhNode *front = LoadNextNode(stream, interior);
			BvhNode *back = LoadNextNode(stream, interior);

			interior->mFront = front;
			interior->mBack = back;

			front->mDepth = interior->mDepth + 1;
			back->mDepth = interior->mDepth + 1;

			tQueue.push(front);			
			tQueue.push(back);
		}
	}


	bvh->mBox = bvh->mRoot->GetBox();

	cout << "... finished loading " << bvh->mNumNodes << " nodes" << endl;
	cout << "scene box: " << bvh->mBox << endl;

	
	///////////
	//-- post process nodes

	bvh->PostProcess();
	
	// set virtual leaves for specified number of triangles
	bvh->SetVirtualLeaves(INITIAL_TRIANGLES_PER_VIRTUAL_LEAVES);

	bvh->UpdateNumLeaves(bvh->mRoot);
	// compute unique ids
	bvh->ComputeIds();
	// specify bounds for occlusion tests
	bvh->RecomputeBounds();

	// compute and print stats
	bvh->ComputeBvhStats();
	bvh->PrintBvhStats();

	return bvh;
}


}
